from mpi4py import MPI
import numpy as np
import matplotlib.pyplot as plt
import time
from funcs import *



comm = MPI.COMM_WORLD
size= comm.Get_size(); me = comm.Get_rank()
pid_main = 0; idle = 0; pia = 0.; n=0; ndom=0


prey_eat = 1 
pred_eat = 10

def iter_step_worker():
### WORKER PROCESS
   
    prey = comm.recv(source=0, tag=11)
    prey_energy = comm.recv(source=0, tag=12)
    prey_sense = comm.recv(source=0, tag=13)
    prey_speed = comm.recv(source=0, tag=14)
    food = comm.recv(source=0, tag=15)
    pred = comm.recv(source=0, tag=16)
    pred_energy = comm.recv(source=0, tag=17)
    pred_sense = comm.recv(source=0, tag=18)
    pred_speed = comm.recv(source=0, tag=19)
    
    ##allocate creatures to worker processes
    l=len(prey)
    l=l//(size-1)
    k=len(pred)
    k=k//(size-1)
    
    if me == size-1:
        my_prey = prey[(me-1)*l:]
        my_energy = prey_energy[(me-1)*l:]
        my_sense = prey_sense[(me-1)*l:]
        my_speed = prey_speed[(me-1)*l:]
        my_pred = pred[(me-1)*k:]
        my_pred_energy = pred_energy[(me-1)*k:]
        my_pred_sense = pred_sense[(me-1)*k:]
        my_pred_speed = pred_speed[(me-1)*k:]
    else:
        my_prey = prey[(me-1)*l:l*me]
        my_energy = prey_energy[(me-1)*l:l*me]
        my_sense = prey_sense[(me-1)*l:l*me]
        my_speed = prey_speed[(me-1)*l:l*me]
        my_pred = pred[(me-1)*k:k*me]
        my_pred_energy = pred_energy[(me-1)*k:k*me]
        my_pred_sense = pred_sense[(me-1)*k:k*me]
        my_pred_speed = pred_speed[(me-1)*k:k*me]
        
    ## for each creature find everything in it's sensing distance     
    prey_want_to_eat = np.zeros((len(my_prey),2))
    pred_want_to_eat = np.zeros((len(my_pred),2))      
    
    for [creature,energy,sense,speed,role] in [[my_prey,my_energy,my_sense,my_speed,"prey"],[my_pred,my_pred_energy,my_pred_sense,my_pred_speed,"pred"]]:

        if len(creature)==0: continue
    	
        near_food =  []
        near_prey = []
        near_pred = []
        if role == 'prey':
            prey_want_to_eat = np.zeros((len(creature),2))
        elif role == 'pred':
            pred_want_to_eat = np.zeros((len(creature),2))
        nearest_prey = np.zeros((len(creature),2))
        nearest_pred = np.zeros((len(creature),2))
    
        if len(creature)>0:
    
            for i,v in enumerate(creature):
                dist_food = np.zeros(food.shape)
                for j in range(len(food)):
                    dist_food[j,0] = np.min((food[j,0] - v[0], lim_x[1]-(food[j,0] - v[0])))
                    dist_food[j,1] = np.min((food[j,1] - v[1], lim_y[1]-(food[j,1] - v[1])))
                dist_food = dist_food**2
                dist_food = dist_food[:,0]+dist_food[:,1]
                dist_food = np.stack((dist_food,np.arange(len(dist_food))),axis=1)
                dist_food = dist_food[dist_food[:, 0].argsort()]
                idx = np.count_nonzero(dist_food[:,0] < sense[i-1]**2)
                near_food.append(dist_food[:idx,1])
                if role == 'prey':
                    prey_want_to_eat[i] = dist_food[0,:]
        	    
                dist_prey = np.zeros(prey.shape)
                for j in range(len(prey)):
                    dist_prey[j,0] = np.min((prey[j,0] - v[0], lim_x[1]-(prey[j,0] - v[0])))
                    dist_prey[j,1] = np.min((prey[j,1] - v[1], lim_y[1]-(prey[j,1] - v[1])))
                dist_prey = dist_prey**2
                dist_prey = dist_prey[:,0]+dist_prey[:,1]
                dist_prey = np.stack((dist_prey,np.arange(len(dist_prey))),axis=1)
                dist_prey = dist_prey[dist_prey[:, 0].argsort()]
                idx = np.count_nonzero(dist_prey[:,0] < sense[i]**2)
                near_prey.append(dist_prey[1:idx,1])
                if role == 'pred':
                    near_prey.append(dist_prey[:idx,1])
                    pred_want_to_eat[i] = dist_prey[0,:]
                dist_pred = np.zeros(pred.shape)
                for j in range(len(pred)):
                    dist_pred[j,0] = np.min((pred[j,0] - v[0], lim_x[1]-(pred[j,0] - v[0])))
                    dist_pred[j,1] = np.min((pred[j,1] - v[1], lim_y[1]-(pred[j,1] - v[1])))
                dist_pred = dist_pred**2
                dist_pred = dist_pred[:,0]+dist_pred[:,1]
                dist_pred = np.stack((dist_pred,np.arange(len(dist_pred))),axis=1)
                dist_pred = dist_pred[dist_pred[:, 0].argsort()]
                idx = np.count_nonzero(dist_pred[:,0] < sense[i]**2)
                near_pred.append(dist_pred[1:idx,1])
                if role == 'prey':
                    near_pred.append(dist_pred[:idx,1])
             
            creature,energy = move(creature,prey,pred,energy,sense,speed,food,near_food,near_prey,near_pred,role)
            
            if role == 'prey': my_prey, my_energy = creature, energy
            elif role == 'pred': my_pred , my_pred_energy = creature, energy
        
    ## send all generated data back	
    comm.send(my_prey, dest=0, tag=21)
    comm.send(my_energy, dest=0, tag=22)
    comm.send(my_sense, dest=0, tag=23)
    comm.send(my_speed, dest=0, tag=24)
    comm.send(my_pred, dest=0, tag=26)
    comm.send(my_pred_energy, dest=0, tag=27)
    comm.send(my_pred_sense, dest=0, tag=28)
    comm.send(my_pred_speed, dest=0, tag=29)
    comm.send(prey_want_to_eat, dest=0, tag=50)
    comm.send(pred_want_to_eat, dest=0, tag=51)
### WORKER PROCESS 

def iter_step_main(gen,prey,prey_energy,prey_sense,prey_speed,food,pred,pred_energy,pred_sense,pred_speed,food_eaten,spawns_food):       
### MAIN PROCESS
    ## send data to sub processes
    for i in range(1,size):
        comm.send(prey,dest=i, tag=11)
        comm.send(prey_energy,dest=i, tag=12)
        comm.send(prey_sense,dest=i, tag=13)
        comm.send(prey_speed,dest=i, tag=14)
        comm.send(food,dest=i, tag=15)
        comm.send(pred,dest=i, tag=16)
        comm.send(pred_energy,dest=i, tag=17)
        comm.send(pred_sense,dest=i, tag=18)
        comm.send(pred_speed,dest=i, tag=19)  
    ## recollect data from worker processes
    buf = np.array([]).reshape((0,2))
    buf2 = np.array([]).reshape((0))
    buf3 = np.array([]).reshape((0))
    buf4 = np.array([]).reshape((0))
    buf6 = np.array([]).reshape((0,2))
    buf7 = np.array([]).reshape((0))
    buf8 = np.array([]).reshape((0))
    buf9 = np.array([]).reshape((0))
    buf50 = np.array([]).reshape((0,2))
    buf51 = np.array([]).reshape((0,2))
    for i in range(1,size):
        prey = comm.recv(source=i, tag=21)
        prey_energy = comm.recv(source=i, tag=22)
        prey_sense = comm.recv(source=i, tag=23)
        prey_speed = comm.recv(source=i, tag=24)
        pred = comm.recv(source=i, tag=26)
        pred_energy = comm.recv(source=i, tag=27)
        pred_sense = comm.recv(source=i, tag=28)
        pred_speed = comm.recv(source=i, tag=29)
        prey_want_to_eat = comm.recv(source=i, tag=50)
        pred_want_to_eat = comm.recv(source=i, tag=51)
        buf = np.concatenate((buf,prey),axis=0)
        buf2 = np.concatenate((buf2,prey_energy),axis=0)
        buf3 = np.concatenate((buf3,prey_sense),axis=0)
        buf4 = np.concatenate((buf4,prey_speed),axis=0)
        buf6 = np.concatenate((buf6,pred),axis=0)
        buf7 = np.concatenate((buf7,pred_energy),axis=0)
        buf8 = np.concatenate((buf8,pred_sense),axis=0)
        buf9 = np.concatenate((buf9,pred_speed),axis=0)
        buf50 = np.concatenate((buf50,prey_want_to_eat),axis=0)
        buf51 = np.concatenate((buf51,pred_want_to_eat),axis=0)
    prey = buf
    prey_energy = buf2
    prey_sense = buf3
    prey_speed = buf4
    pred = buf6
    pred_energy = buf7
    pred_sense = buf8
    pred_speed = buf9
    prey_want_to_eat = buf50
    pred_want_to_eat = buf51
    
    ## prey and predator eat
    want_to_eat = [prey_want_to_eat,pred_want_to_eat]
    eat = [prey_eat,pred_eat]
    energy = [prey_energy, pred_energy]
    prey_eaten = 0
    food_eaten_it = 0
    for j in [0,1]:      
        can_eat = want_to_eat[j][:,0]<eat[j]**2
        can_eat_idx = np.arange(len(want_to_eat[j]))[can_eat]
        already_eaten=[]

        for i in can_eat_idx:
            if want_to_eat[j][i,1] in already_eaten:
                continue
            count = np.count_nonzero(want_to_eat[j][i,1]==want_to_eat[j][can_eat_idx,1])
            if count == 1:
                if j == 0 and food[int(want_to_eat[j][i,1])][0] != np.inf:
                    food[int(want_to_eat[j][i,1])] = np.array([np.inf,np.inf])
                    prey_energy[i] += 100
                    food_eaten_it += 1
                elif j == 1 and prey_energy[int(want_to_eat[j][i,1])] > 0:
                    prey_energy[int(want_to_eat[j][i,1])] = 0
                    pred_energy[i] += 150
                    prey_eaten += 1

            else:
                decide = np.random.randint(count)
                decide = np.nonzero(want_to_eat[j][i,1]==want_to_eat[j][can_eat_idx,1])[0][decide]
                if j == 0 and food[int(want_to_eat[j][int(decide),1])][0] != np.inf:
                    food[int(want_to_eat[j][int(decide),1])] = np.array([np.inf,np.inf])
                    prey_energy[decide] += 100
                    food_eaten_it += 1
                elif j == 1 and prey_energy[int(want_to_eat[j][i,1])] > 0 :
                    prey_energy[int(want_to_eat[j][i,1])] = 0
                    pred_energy[i] = prey_energy[int(want_to_eat[j][i,1])] + 80
                    #pred_energy[i] += 150
                    prey_eaten += 1
                already_eaten.append(want_to_eat[j][decide,1])
   
    food = np.delete(food,np.nonzero(food[:,0] == np.inf),axis=0)

   
    ## update food and creatures
    prey,prey_energy,prey_sense,prey_speed = check_alive(prey,prey_energy,prey_sense,prey_speed,)
    pred,pred_energy,pred_sense,pred_speed = check_alive(pred,pred_energy,pred_sense,pred_speed,)

    food_eaten = np.concatenate((food_eaten[1:10], [food_eaten_it]))
    food = spawn_food(food,spawns_food,food_eaten[0])


    prey,prey_energy,prey_sense,prey_speed = reproduce(prey,prey_energy,prey_sense,prey_speed,'prey')
    pred,pred_energy,pred_sense,pred_speed = reproduce(pred,pred_energy,pred_sense,pred_speed,'pred')

    return prey,prey_energy,prey_sense,prey_speed,food,pred,pred_energy,pred_sense,pred_speed,food_eaten
